@@ -20,7 +20,11 @@ from mellea.backends import BaseModelSubclass
 from mellea.backends.aloras import Alora, AloraBackendMixin
 from mellea.backends.formatter import Formatter, FormatterBackend, TemplateFormatter
 from mellea.backends.model_ids import ModelIdentifier
-from mellea.backends.tools import convert_tools_to_json, get_tools_from_action
+from mellea.backends.tools import (
+    add_tools_from_context_actions,
+    add_tools_from_model_options,
+    convert_tools_to_json,
+)
 from mellea.backends.types import ModelOption
 from mellea.helpers.fancy_logger import FancyLogger
 from mellea.stdlib.base import (
@@ -30,7 +34,6 @@ from mellea.stdlib.base import (
     GenerateLog,
     ModelOutputThunk,
     ModelToolCall,
-    TemplateRepresentation,
 )
 from mellea.stdlib.chat import Message
 from mellea.stdlib.requirement import ALoraRequirement, LLMaJRequirement, Requirement
@@ -114,7 +117,6 @@ class OpenAIBackend(FormatterBackend, AloraBackendMixin):
         self.from_mellea_model_opts_map_chats = {
             ModelOption.SEED: "seed",
             ModelOption.MAX_NEW_TOKENS: "max_completion_tokens",
-            ModelOption.THINKING: "reasoning_effort",
         }
 
         # See notes above.
@@ -348,6 +350,40 @@ class OpenAIBackend(FormatterBackend, AloraBackendMixin):
             ),
         )
 
+    @staticmethod
+    def message_to_openai_message(msg: Message):
+        if msg.images is not None:
+            img_list = [
+                {
+                    "type": "image_url",
+                    "image_url": {"url": f"data:image/png;base64,{img}"},
+                }
+                for img in msg.images
+            ]
+
+            return {
+                "role": msg.role,
+                "content": [{"type": "text", "text": msg.content}, *img_list],
+            }
+        else:
+            return {"role": msg.role, "content": msg.content}
+            # Target format:
+            # {
+            #     "role": "user",
+            #     "content": [
+            #       {
+            #         "type": "text",
+            #         "text": "What's in this picture?"
+            #       },
+            #       {
+            #         "type": "image_url",
+            #         "image_url": {
+            #           "url": "data:image/jpeg;base64,<base64_string>"
+            #         }
+            #       }
+            #     ]
+            #   }
+
     def _generate_from_chat_context_standard(
         self,
         action: Component | CBlock,
@@ -382,7 +418,7 @@ class OpenAIBackend(FormatterBackend, AloraBackendMixin):
         system_prompt = model_opts.get(ModelOption.SYSTEM_PROMPT, "")
         if system_prompt != "":
             conversation.append({"role": "system", "content": system_prompt})
-        conversation.extend([{"role": m.role, "content": m.content} for m in messages])
+        conversation.extend([self.message_to_openai_message(m) for m in messages])
 
         if format is not None:
             response_format = {
@@ -404,28 +440,13 @@ class OpenAIBackend(FormatterBackend, AloraBackendMixin):
                     f"Tool calling typically uses constrained generation, but you have specified a `format` in your generate call. NB: tool calling is superseded by format; we will NOT call tools for your request: {action}"
                 )
             else:
-                if isinstance(action, Component) and isinstance(
-                    action.format_for_llm(), TemplateRepresentation
-                ):
-                    tools = get_tools_from_action(action)
+                add_tools_from_model_options(tools, model_opts)
+                add_tools_from_context_actions(tools, ctx.actions_for_available_tools())
 
-                model_options_tools = model_opts.get(ModelOption.TOOLS, None)
-                if model_options_tools is not None:
-                    assert isinstance(model_options_tools, dict)
-                    for fn_name in model_options_tools:
-                        # invariant re: relationship between the model_options set of tools and the TemplateRepresentation set of tools
-                        assert fn_name not in tools.keys(), (
-                            f"Cannot add tool {fn_name} because that tool was already defined in the TemplateRepresentation for the action."
-                        )
-                        # type checking because ModelOptions is an untyped dict and the calling convention for tools isn't clearly documented at our abstraction boundaries.
-                        assert type(fn_name) is str, (
-                            "When providing a `ModelOption.TOOLS` parameter to `model_options`, always used the type Dict[str, Callable] where `str` is the function name and the callable is the function."
-                        )
-                        assert callable(model_options_tools[fn_name]), (
-                            "When providing a `ModelOption.TOOLS` parameter to `model_options`, always used the type Dict[str, Callable] where `str` is the function name and the callable is the function."
-                        )
-                        # Add the model_options tool to the existing set of tools.
-                        tools[fn_name] = model_options_tools[fn_name]
+                # Add the tools from the action for this generation last so that
+                # they overwrite conflicting names.
+                add_tools_from_context_actions(tools, [action])
+            FancyLogger.get_logger().info(f"Tools for call: {tools.keys()}")
 
         thinking = model_opts.get(ModelOption.THINKING, None)
         if type(thinking) is bool and thinking:
@@ -433,15 +454,14 @@ class OpenAIBackend(FormatterBackend, AloraBackendMixin):
             thinking = "medium"
 
         formatted_tools = convert_tools_to_json(tools)
+        use_tools = len(formatted_tools) > 0
+
         chat_response: ChatCompletion = self._client.chat.completions.create(
             model=self._hf_model_id,
             messages=conversation,  # type: ignore
             reasoning_effort=thinking,  # type: ignore
             response_format=response_format,  # type: ignore
-            tool_choice=(
-                "auto" if formatted_tools and len(formatted_tools) > 0 else "none"
-            ),
-            tools=formatted_tools,  # type: ignore
+            tools=formatted_tools if use_tools else None,  # type: ignore
             # parallel_tool_calls=False, # We only support calling one tool per turn. But we do the choosing on our side so we leave this False.
             **self._make_backend_specific_and_remove(
                 model_opts, is_chat_context=ctx.is_chat_context
