@@ -3,12 +3,17 @@
 from __future__ import annotations
 
 import abc
+import base64
+import binascii
 import datetime
 from collections.abc import Callable, Iterable, Mapping
 from copy import deepcopy
 from dataclasses import dataclass
+from io import BytesIO
 from typing import Any, Protocol, runtime_checkable
 
+from PIL import Image as PILImage
+
 from mellea.helpers.fancy_logger import FancyLogger
 
 
@@ -43,6 +48,70 @@ class CBlock:
         return f"CBlock({self.value}, {self._meta.__repr__()})"
 
 
+class ImageBlock:
+    """A `ImageBlock` represents an image (as base64 PNG)."""
+
+    def __init__(self, value: str, meta: dict[str, Any] | None = None):
+        """Initializes the ImageBlock with a base64 PNG string representation and some metadata."""
+        assert self.is_valid_base64_png(value), (
+            "Invalid base64 string representation of image."
+        )
+        self._value = value
+        self._meta = {} if meta is None else meta
+
+    @staticmethod
+    def is_valid_base64_png(s: str) -> bool:
+        """Checks if a string is a valid base64 string [AIA PAI Nc Hin R v1.0]."""
+        try:
+            # Check if the string has a data URI prefix and remove it.
+            if "data:" in s and "base64," in s:
+                s = s.split("base64,")[1]
+
+            # Add padding if necessary
+            s = s.strip()
+            mod4 = len(s) % 4
+            if mod4 > 0:
+                s = s + "=" * (4 - mod4)
+
+            # Attempt to decode the Base64 string
+            decoded_data = base64.b64decode(s, validate=True)
+
+            # The official PNG signature is 8 bytes long.
+            png_signature = b"\x89PNG\r\n\x1a\n"
+
+            if decoded_data.startswith(png_signature):
+                return True
+            else:
+                return False
+
+            return True
+        except (binascii.Error, ValueError):
+            return False
+
+    @staticmethod
+    def pil_to_base64(image: PILImage.Image) -> str:
+        """Converts a PIL image to a base64 string representation."""
+        img_io = BytesIO()
+        image.save(img_io, "PNG")
+        return base64.b64encode(img_io.getvalue()).decode("utf-8")
+
+    @classmethod
+    def from_pil_image(
+        cls, image: PILImage.Image, meta: dict[str, Any] | None = None
+    ) -> ImageBlock:
+        """Converts a PIL image to a base64 string representation."""
+        image_base64 = cls.pil_to_base64(image)
+        return cls(image_base64, meta)
+
+    def __str__(self):
+        """Stringifies the block."""
+        return self._value
+
+    def __repr__(self):
+        """Provides a python-parsable representation of the block (usually)."""
+        return f"ImageBlock({self._value}, {self._meta.__repr__()})"
+
+
 @runtime_checkable
 class Component(Protocol):
     """A `Component` is a composite data structure that is intended to be represented to an LLM."""
@@ -59,6 +128,25 @@ class Component(Protocol):
         raise NotImplementedError("format_for_llm isn't implemented by default")
 
 
+def get_images_from_component(c: Component) -> None | list[ImageBlock]:
+    """Gets images from a `Component` if they are present and a non-empty list, otherwise returns None."""
+    if hasattr(c, "images"):
+        imgs = c.images
+        if imgs is not None:
+            assert isinstance(imgs, list), "images field must be a list."
+            assert all(isinstance(im, ImageBlock) for im in imgs), (
+                "all elements of images list must be ImageBlocks."
+            )
+            if len(imgs) == 0:
+                return None
+            else:
+                return imgs
+        else:
+            return None
+    else:
+        return None
+
+
 class ModelOutputThunk(CBlock):
     """A `ModelOutputThunk` is a special type of `CBlock` that we know came from a model's output. It is possible to instantiate one without the output being computed yet."""
 
@@ -161,6 +249,14 @@ class Context(abc.ABC):
         """Provides a linear list of context components to use for generation, or None if that is not possible to construct."""
         ...
 
+    @abc.abstractmethod
+    def actions_for_available_tools(self) -> list[Component | CBlock] | None:
+        """Provides a list of actions to extract tools from for use with during generation, or None if that's not possible.
+
+        Can be used to make the available tools differ from the tools of all the actions in the context.
+        """
+        ...
+
     @abc.abstractmethod
     def full_event_log(self) -> list[Component | CBlock]:
         """Provides a list of all events stored in the context."""
@@ -210,6 +306,14 @@ class BasicContext(Context, abc.ABC):
         self._ctx = []
         self._log_ctx = []
 
+    def actions_for_available_tools(self) -> list[Component | CBlock] | None:
+        """Provides a list of actions to extract tools from for use with during generation, or None if that's not possible.
+
+        Can be used to make the available tools differ from the tools of all the actions in the context.
+        In most cases, this will just be the same context as `render_for_generation`.
+        """
+        return self.render_for_generation()
+
     def last_output(self):
         """The last output thunk of the context."""
         for c in self._ctx[::-1]:
@@ -436,6 +540,7 @@ class TemplateRepresentation:
     fields: list[Any] | None = None
     template: str | None = None
     template_order: list[str] | None = None
+    images: list[ImageBlock] | None = None
 
 
 @dataclass
