@@ -3,10 +3,10 @@
 from __future__ import annotations
 
 import contextvars
-from collections.abc import Generator
-from contextlib import contextmanager
 from copy import deepcopy
-from typing import Any, Literal, Optional
+from typing import Any, Literal, overload
+
+from PIL import Image as PILImage
 
 from mellea.backends import Backend, BaseModelSubclass
 from mellea.backends.formatter import FormatterBackend
@@ -24,6 +24,7 @@ from mellea.stdlib.base import (
     Context,
     ContextTurn,
     GenerateLog,
+    ImageBlock,
     LinearContext,
     ModelOutputThunk,
     SimpleContext,
@@ -69,12 +70,16 @@ def backend_name_to_class(name: str) -> Any:
         from mellea.backends.watsonx import WatsonxAIBackend
 
         return WatsonxAIBackend
+    elif name == "litellm":
+        from mellea.backends.litellm import LiteLLMBackend
+
+        return LiteLLMBackend
     else:
         return None
 
 
 def start_session(
-    backend_name: Literal["ollama", "hf", "openai", "watsonx"] = "ollama",
+    backend_name: Literal["ollama", "hf", "openai", "watsonx", "litellm"] = "ollama",
     model_id: str | ModelIdentifier = IBM_GRANITE_3_3_8B,
     ctx: Context | None = SimpleContext(),
     *,
@@ -224,16 +229,204 @@ class MelleaSession:
         self.reset()
         self._backend_stack.clear()
         if hasattr(self.backend, "close"):
-            self.backend.close()
+            self.backend.close()  # type: ignore
 
     def summarize(self) -> ModelOutputThunk:
         """Summarizes the current context."""
         raise NotImplementedError()
 
+    @staticmethod
+    def _parse_and_clean_image_args(
+        images_: list[ImageBlock] | list[PILImage.Image] | None = None,
+    ) -> list[ImageBlock] | None:
+        images: list[ImageBlock] | None = None
+        if images_ is not None:
+            assert isinstance(images_, list), "Images should be a list or None."
+
+            if len(images_) > 0:
+                if isinstance(images_[0], PILImage.Image):
+                    images = [
+                        ImageBlock.from_pil_image(i)
+                        for i in images_
+                        if isinstance(i, PILImage.Image)
+                    ]
+                else:
+                    images = images_  # type: ignore
+                assert isinstance(images, list)
+                assert all(isinstance(i, ImageBlock) for i in images), (
+                    "All images should be ImageBlocks now."
+                )
+            else:
+                images = None
+        return images
+
+    @overload
+    def act(
+        self,
+        action: Component,
+        *,
+        strategy: SamplingStrategy | None = None,
+        return_sampling_results: Literal[False] = False,
+        format: type[BaseModelSubclass] | None = None,
+        model_options: dict | None = None,
+        tool_calls: bool = False,
+    ) -> ModelOutputThunk: ...
+
+    @overload
+    def act(
+        self,
+        action: Component,
+        *,
+        strategy: SamplingStrategy | None = None,
+        return_sampling_results: Literal[True],
+        format: type[BaseModelSubclass] | None = None,
+        model_options: dict | None = None,
+        tool_calls: bool = False,
+    ) -> SamplingResult: ...
+
+    def act(
+        self,
+        action: Component,
+        *,
+        requirements: list[Requirement] | None = None,
+        strategy: SamplingStrategy | None = None,
+        return_sampling_results: bool = False,
+        format: type[BaseModelSubclass] | None = None,
+        model_options: dict | None = None,
+        tool_calls: bool = False,
+    ) -> ModelOutputThunk | SamplingResult:
+        """Runs a generic action, and adds both the action and the result to the context.
+
+        Args:
+            action: the Component from which to generate.
+            requirements: used as additional requirements when a sampling strategy is provided
+            strategy: a SamplingStrategy that describes the strategy for validating and repairing/retrying for the instruct-validate-repair pattern. None means that no particular sampling strategy is used.
+            return_sampling_results: attach the (successful and failed) sampling attempts to the results.
+            format: if set, the BaseModel to use for constrained decoding.
+            model_options: additional model options, which will upsert into the model/backend's defaults.
+            tool_calls: if true, tool calling is enabled.
+
+        Returns:
+            A ModelOutputThunk if `return_sampling_results` is `False`, else returns a `SamplingResult`.
+        """
+        sampling_result: SamplingResult | None = None
+        generate_logs: list[GenerateLog] = []
+
+        if return_sampling_results:
+            assert strategy is not None, (
+                "Must provide a SamplingStrategy when return_sampling_results==True"
+            )
+
+        if strategy is None:
+            result = self.backend.generate_from_context(
+                action,
+                ctx=self.ctx,
+                format=format,
+                model_options=model_options,
+                generate_logs=generate_logs,
+                tool_calls=tool_calls,
+            )
+            assert len(generate_logs) == 1, "Simple call can only add one generate_log"
+            generate_logs[-1].is_final_result = True
+
+        else:
+            # Default validation strategy just validates all of the provided requirements.
+            if strategy.validate is None:
+                strategy.validate = lambda reqs, val_ctx, output: self.validate(
+                    reqs, output=output
+                )
+
+            # Default generation strategy just generates from context.
+            if strategy.generate is None:
+                strategy.generate = (
+                    lambda sample_action,
+                    gen_ctx,
+                    g_logs: self.backend.generate_from_context(
+                        sample_action,
+                        ctx=gen_ctx,
+                        format=format,
+                        model_options=model_options,
+                        generate_logs=g_logs,
+                        tool_calls=tool_calls,
+                    )
+                )
+
+            if requirements is None:
+                requirements = []
+
+            sampling_result = strategy.sample(
+                action, self.ctx, requirements=requirements, generate_logs=generate_logs
+            )
+
+            # make sure that one Log is marked as the one related to sampling_result.result
+            if sampling_result.success:
+                # if successful, the last log is the one related
+                generate_logs[-1].is_final_result = True
+            else:
+                # Find the log where log.result and sampling_result.result match
+                selected_log = [
+                    log for log in generate_logs if log.result == sampling_result.result
+                ]
+                assert len(selected_log) == 1, (
+                    "There should only be exactly one log corresponding to the single result. "
+                )
+                selected_log[0].is_final_result = True
+
+            result = sampling_result.result
+
+        self.ctx.insert_turn(ContextTurn(action, result), generate_logs=generate_logs)
+
+        if return_sampling_results:
+            assert (
+                sampling_result is not None
+            )  # Needed for the type checker but should never happen.
+            return sampling_result
+        else:
+            return result
+
+    @overload
     def instruct(
         self,
         description: str,
         *,
+        images: list[ImageBlock] | list[PILImage.Image] | None = None,
+        requirements: list[Requirement | str] | None = None,
+        icl_examples: list[str | CBlock] | None = None,
+        grounding_context: dict[str, str | CBlock | Component] | None = None,
+        user_variables: dict[str, str] | None = None,
+        prefix: str | CBlock | None = None,
+        output_prefix: str | CBlock | None = None,
+        strategy: SamplingStrategy | None = None,
+        return_sampling_results: Literal[False] = False,
+        format: type[BaseModelSubclass] | None = None,
+        model_options: dict | None = None,
+        tool_calls: bool = False,
+    ) -> ModelOutputThunk: ...
+
+    @overload
+    def instruct(
+        self,
+        description: str,
+        *,
+        images: list[ImageBlock] | list[PILImage.Image] | None = None,
+        requirements: list[Requirement | str] | None = None,
+        icl_examples: list[str | CBlock] | None = None,
+        grounding_context: dict[str, str | CBlock | Component] | None = None,
+        user_variables: dict[str, str] | None = None,
+        prefix: str | CBlock | None = None,
+        output_prefix: str | CBlock | None = None,
+        strategy: SamplingStrategy | None = None,
+        return_sampling_results: Literal[True],
+        format: type[BaseModelSubclass] | None = None,
+        model_options: dict | None = None,
+        tool_calls: bool = False,
+    ) -> SamplingResult: ...
+
+    def instruct(
+        self,
+        description: str,
+        *,
+        images: list[ImageBlock] | list[PILImage.Image] | None = None,
         requirements: list[Requirement | str] | None = None,
         icl_examples: list[str | CBlock] | None = None,
         grounding_context: dict[str, str | CBlock | Component] | None = None,
@@ -261,11 +454,16 @@ class MelleaSession:
             format: If set, the BaseModel to use for constrained decoding.
             model_options: Additional model options, which will upsert into the model/backend's defaults.
             tool_calls: If true, tool calling is enabled.
+            images: A list of images to be used in the instruction or None if none.
         """
+
         requirements = [] if requirements is None else requirements
         icl_examples = [] if icl_examples is None else icl_examples
         grounding_context = dict() if grounding_context is None else grounding_context
-        # all instruction options are forwarded to create a new Instruction object
+
+        images = self._parse_and_clean_image_args(images)
+
+        # All instruction options are forwarded to create a new Instruction object.
         i = Instruction(
             description=description,
             requirements=requirements,
@@ -274,73 +472,25 @@ class MelleaSession:
             user_variables=user_variables,
             prefix=prefix,
             output_prefix=output_prefix,
+            images=images,
         )
 
-        res = None
-        generate_logs: list[GenerateLog] = []
-        if strategy is None:
-            result = self.backend.generate_from_context(
-                i,
-                ctx=self.ctx,
-                format=format,
-                model_options=model_options,
-                generate_logs=generate_logs,
-                tool_calls=tool_calls,
-            )
-
-            # make sure that one Log is marked as the one related to result
-            assert len(generate_logs) == 1, "Simple call can only add one generate_log"
-            generate_logs[0].is_final_result = True
-        else:
-            if strategy.validate is None:
-                strategy.validate = lambda reqs, output: self.validate(  # type: ignore
-                    reqs,
-                    output=output,  # type: ignore
-                )  # type: ignore
-            if strategy.generate is None:
-                strategy.generate = (
-                    lambda instruction, g_logs: self.backend.generate_from_context(
-                        instruction,
-                        ctx=self.ctx,
-                        format=format,
-                        model_options=model_options,
-                        generate_logs=g_logs,
-                        tool_calls=tool_calls,
-                    )
-                )
-
-            # sample
-            res = strategy.sample(i, generate_logs=generate_logs)
-
-            # make sure that one Log is marked as the one related to res.result
-            if res.success:
-                # if successful, the last log is the one related
-                generate_logs[-1].is_final_result = True
-            else:
-                # find the one where log.result and res.result match
-                selected_log = [
-                    log for log in generate_logs if log.result == res.result
-                ]
-                assert len(selected_log) == 1, (
-                    "There should only be exactly one log corresponding to the single result. "
-                )
-                selected_log[0].is_final_result = True
-
-            result = res.result
-
-        self.ctx.insert_turn(ContextTurn(i, result), generate_logs=generate_logs)
-
-        if return_sampling_results:
-            assert res is not None, "Asking for sampling results without sampling."
-            return res
-        else:
-            return result
+        return self.act(
+            i,
+            requirements=i.requirements,
+            strategy=strategy,
+            return_sampling_results=return_sampling_results,
+            format=format,
+            model_options=model_options,
+            tool_calls=tool_calls,
+        )  # type: ignore[call-overload]
 
     def chat(
         self,
         content: str,
         role: Message.Role = "user",
         *,
+        images: list[ImageBlock] | list[PILImage.Image] | None = None,
         user_variables: dict[str, str] | None = None,
         format: type[BaseModelSubclass] | None = None,
         model_options: dict | None = None,
@@ -353,36 +503,20 @@ class MelleaSession:
             )
         else:
             content_resolved = content
-        user_message = Message(role=role, content=content_resolved)
-        generate_logs: list[GenerateLog] = []
-        output_thunk = self.backend.generate_from_context(
-            action=user_message,
-            ctx=self.ctx,
+        images = self._parse_and_clean_image_args(images)
+        user_message = Message(role=role, content=content_resolved, images=images)
+
+        result = self.act(
+            user_message,
             format=format,
             model_options=model_options,
-            generate_logs=generate_logs,
             tool_calls=tool_calls,
         )
-        # make sure that the last and only Log is marked as the one related to result
-        assert len(generate_logs) == 1, "Simple call can only add one generate_log"
-        generate_logs[0].is_final_result = True
+        parsed_assistant_message = result.parsed_repr
+        assert isinstance(parsed_assistant_message, Message)
 
-        parsed_assistant_message = output_thunk.parsed_repr
-        assert type(parsed_assistant_message) is Message
-        self.ctx.insert_turn(
-            ContextTurn(user_message, output_thunk), generate_logs=generate_logs
-        )
         return parsed_assistant_message
 
-    def act(self, c: Component, tool_calls: bool = False) -> Any:
-        """Runs a generic action, and adds both the action and the result to the context."""
-        generate_logs: list[GenerateLog] = []
-        result: ModelOutputThunk = self.backend.generate_from_context(
-            c, self.ctx, generate_logs=generate_logs, tool_calls=tool_calls
-        )
-        self.ctx.insert_turn(turn=ContextTurn(c, result), generate_logs=generate_logs)
-        return result
-
     def validate(
         self,
         reqs: Requirement | list[Requirement],
@@ -414,65 +548,6 @@ class MelleaSession:
 
         return rvs
 
-    def req(self, *args, **kwargs):
-        """Shorthand for Requirement.__init__(...)."""
-        return req(*args, **kwargs)
-
-    def check(self, *args, **kwargs):
-        """Shorthand for Requirement.__init__(..., check_only=True)."""
-        return check(*args, **kwargs)
-
-    def load_default_aloras(self):
-        """Loads the default Aloras for this model, if they exist and if the backend supports."""
-        from mellea.backends.huggingface import LocalHFBackend
-
-        if self.backend.model_id == IBM_GRANITE_3_2_8B and isinstance(
-            self.backend, LocalHFBackend
-        ):
-            from mellea.backends.aloras.huggingface.granite_aloras import (
-                add_granite_aloras,
-            )
-
-            add_granite_aloras(self.backend)
-            return
-        self._session_logger.warning(
-            "This model/backend combination does not support any aloras."
-        )
-
-    def genslot(
-        self,
-        gen_slot: Component,
-        model_options: dict | None = None,
-        format: type[BaseModelSubclass] | None = None,
-        tool_calls: bool = False,
-    ) -> ModelOutputThunk:
-        """Call generative Slot on a GenerativeSlot Component.
-
-        Args:
-            gen_slot (GenerativeSlot Component): A generative slot
-
-        Returns:
-            ModelOutputThunk: Output thunk
-        """
-        generate_logs: list[GenerateLog] = []
-        result: ModelOutputThunk = self.backend.generate_from_context(
-            action=gen_slot,
-            ctx=self.ctx,
-            model_options=model_options,
-            format=format,
-            generate_logs=generate_logs,
-            tool_calls=tool_calls,
-        )
-        # make sure that the last and only Log is marked as the one related to result
-        assert len(generate_logs) == 1, "Simple call can only add one generate_log"
-        generate_logs[0].is_final_result = True
-
-        self.ctx.insert_turn(
-            ContextTurn(deepcopy(gen_slot), result), generate_logs=generate_logs
-        )
-
-        return result
-
     def query(
         self,
         obj: Any,
@@ -500,32 +575,9 @@ class MelleaSession:
         assert isinstance(obj, MObjectProtocol)
         q = obj.get_query_object(query)
 
-        generate_logs: list[GenerateLog] = []
-        answer = self.backend.generate_from_context(
-            q,
-            self.ctx,
-            format=format,
-            model_options=model_options,
-            generate_logs=generate_logs,
-            tool_calls=tool_calls,
+        answer = self.act(
+            q, format=format, model_options=model_options, tool_calls=tool_calls
         )
-        # make sure that the last and only Log is marked as the one related to result
-        assert len(generate_logs) == 1, "Simple call can only add one generate_log"
-        generate_logs[0].is_final_result = True
-
-        if isinstance(self.ctx, SimpleContext):
-            self.ctx.insert_turn(ContextTurn(q, answer), generate_logs=generate_logs)
-        elif isinstance(self.ctx, LinearContext) and len(self.ctx._ctx) == 0:
-            FancyLogger.get_logger().info(
-                "Adding the Object Query and its answer as first turn to a Linear Context (Chat History). "
-                "You can now run more .chat() or .instruct() with the object as reference."
-            )
-            self.ctx.insert_turn(ContextTurn(q, answer), generate_logs=generate_logs)
-        else:
-            FancyLogger.get_logger().info(
-                "The Linear Context has not been modified by this query."
-            )
-
         return answer
 
     def transform(
@@ -553,43 +605,12 @@ class MelleaSession:
         assert isinstance(obj, MObjectProtocol)
         t = obj.get_transform_object(transformation)
 
-        generate_logs: list[GenerateLog] = []
-
         # Check that your model / backend supports tool calling.
         # This might throw an error when tools are provided but can't be handled by one or the other.
-        transformed = self.backend.generate_from_context(
-            t,
-            self.ctx,
-            format=format,
-            model_options=model_options,
-            generate_logs=generate_logs,
-            tool_calls=True,
+        transformed = self.act(
+            t, format=format, model_options=model_options, tool_calls=True
         )
 
-        assert len(generate_logs) == 1, "Simple call can only add one generate_log"
-        generate_logs[0].is_final_result = True
-
-        # Insert the new turn into the context. Tool calls are handled afterwards.
-        insert = False
-        if isinstance(self.ctx, SimpleContext):
-            insert = True
-            self.ctx.insert_turn(
-                ContextTurn(t, transformed), generate_logs=generate_logs
-            )
-        elif isinstance(self.ctx, LinearContext) and len(self.ctx._ctx) == 0:
-            insert = True
-            FancyLogger.get_logger().info(
-                "Adding the Object Transform and its result as first turn to a Linear Context (Chat History). "
-                "You can now run more .chat() or .instruct() with the object as reference."
-            )
-            self.ctx.insert_turn(
-                ContextTurn(t, transformed), generate_logs=generate_logs
-            )
-        else:
-            FancyLogger.get_logger().info(
-                "The Linear Context has not been modified by this query."
-            )
-
         tools = self._call_tools(transformed)
 
         # Transform only supports calling one tool call since it cannot currently synthesize multiple outputs.
@@ -609,7 +630,8 @@ class MelleaSession:
                 chosen_tool = tools[0]
 
             FancyLogger.get_logger().warning(
-                f"multiple tool calls returned in transform of {obj} with description '{transformation}'; picked `{chosen_tool.name}`"  # type: ignore
+                f"multiple tool calls returned in transform of {obj} with description '{transformation}'; picked `{chosen_tool.name}`"
+                # type: ignore
             )
 
         if chosen_tool:
@@ -618,11 +640,11 @@ class MelleaSession:
                 FancyLogger.get_logger().warning(
                     f"the transform of {obj} with transformation description '{transformation}' resulted in a tool call with no generated arguments; consider calling the function `{chosen_tool._tool.name}` directly"
                 )
-            if insert:
-                self.ctx.insert(chosen_tool)
-                FancyLogger.get_logger().warning(
-                    "added a tool message from transform to the context as well."
-                )
+
+            self.ctx.insert(chosen_tool)
+            FancyLogger.get_logger().info(
+                "added a tool message from transform to the context"
+            )
             return chosen_tool._tool_output
 
         return transformed
